import numpy as np
import scipy as sc

from .utils import *
from .operators import *

def vn_povm(H, spectrum=False):
    r"""
    Returns a P(O)VM corresponding to the eigenstates of a Hermitian operator.
    """
    L, V = np.linalg.eig(H)
    E = np.array([np.outer(v, v.conj()) for v in V.T])
    return L, E if spectrum else E

def tighten(R):
    r"""
    Tightens frame.
    """
    return sc.linalg.polar(R)[0]

def frame_povm(R):
    r"""
    Lifts tight frame to POVM.
    """
    return np.array([np.outer(r, r.conj()) for r in R.T])

def squish(E):
    S = sc.linalg.fractional_matrix_power(sum(E), -1/2)
    return np.array([S @ e @ S for e in E])

def complete(E):
    d = E.shape[-1]
    return np.vstack([E, (np.eye(d) - sum(E)).reshape(1,d,d)])

def dilate(E):
    EE = []
    mapping = {}
    for i, e in enumerate(E):
        L, V = np.linalg.eig(e)
        mapping[i] = []
        for j in range(len(L)):
            if not np.isclose(L[j], 0):
                EE.append(L[j]*np.outer(V[j], V[j].conj()))
                mapping[i].append(len(EE)-1)
    return np.array(EE)

def coarse_grain(E, mapping):
    return np.array([sum([E[v] for v in V]) for k, V in mapping.items()])

def sample_from_povm(E, rho, n=1):
    p = np.array([(e@rho).trace() for e in E]).real
    return np.random.choice(list(range(len(E))), size=n, p=p)

def implement_povm(E):
    n, d = len(E), E[0].shape[0]
    V = sum([np.kron(basis(n, i), sc.linalg.sqrtm(E[i])) for i in range(n)]).T
    Q, R = np.linalg.qr(V, mode="complete")
    return -Q.conj()

def discriminator_povm(a, b):
    r"""
    Returns a non informationally complete POVM which has the special property
    of distinguishing between two arbitrary states $\mid a \rangle$ and $\mid b\rangle$, which are not necessarily orthogonal (which is impossible with a standard PVM).

    It has three elements:

    $$ \hat{F}_{a} = \frac{1}{1+\mid\langle a \mid b \rangle\mid}(\hat{I} - \mid b \rangle \langle b \mid) $$
    $$ \hat{F}_{b} = \frac{1}{1+\mid\langle a \mid b \rangle\mid}(\hat{I} - \mid a \rangle \langle a \mid) $$
    $$ \hat{F}_{?} = \hat{I} - \hat{F}_{a} - \hat{F}_{b} $$

    The first tests for "not B", the second tests for "not A", and the third outcome represents an inconclusive result.
    """
    d = a.shape[0]
    p = abs(a.conj().T @ b)
    Fa = (1/(1+p))*(np.eye(d) - b @ b.conj().T)
    Fb = (1/(1+p))*(np.eye(d) - a @ a.conj().T)
    Fq = np.eye(d) - Fa - Fb
    return np.array([Fa, Fb, Fq])

#!
def quantumness(A, B=None, p=2):
    B = B if type(B) != type(None) else A
    A, B = Operators(A), Operators(B)
    return pnorm(np.eye(len(A)) - (~A|B), p)

#!
def frame_quantumness(R, S=None, p=2):
    d, n = R.shape
    S = R if type(S) == type(None) else S
    P = np.abs(R.conj().T @ (np.tile(1/np.linalg.norm(S, axis=0), (d, 1))*S))**2
    return pnorm(np.eye(n) - spectral_inverse(P), p)

#!
def weighted_frame_potential(R, t):
    d, n = R.shape
    N = [np.linalg.norm(r) for r in R.T]
    S = np.array([r/N[i] for i, r in enumerate(R.T)])
    w = [(N[i]**2)/d for i, r in enumerate(R.T)]
    return sum([w[i]*w[j]*abs(np.vdot(S[i], S[j]))**(2*t) for j in range(n) for i in range(n)])

#!
def minimum_real_frame_potential(d, t):
    return np.prod(np.array(list(range(1, 2*t, 2))))/\
           np.prod(np.array(list(range(d, d + 2*t,2))))

#!
def minimum_complex_frame_potential(d, t):
    return 1/sc.special.binom(d+t-1, t)


import numpy as np

from .utils import *
from .povm_utils import *

def clock(d):
    r"""
    The clock operator $\hat{Z}$ for dimension $d$.
    """
    w = np.exp(2*np.pi*1j/d)
    return np.diag([w**i for i in range(d)])

def shift(d):
    r"""
    The shift operator $\hat{X}$ for dimension $d$.
    """
    return sum([np.outer(basis(d, i+1), basis(d, i))\
                    if i != d-1 else np.outer(basis(d, 0),basis(d, i))\
                        for i in range(d) for j in range(d)])/d

def discrete_Q(d):
    return d*sc.linalg.logm(clock(d))/(2*np.pi*1j)

def discrete_P(d):
    fft = fft_matrix(d)
    return fft @ discrete_Q(d) @ fft.conj().T 

def displace(d, q, p):
    Z, X = clock(d), shift(d)
    return (-np.exp(1j*np.pi/d))**(q*p)*np.linalg.matrix_power(X,q) @ np.linalg.matrix_power(Z,p)

def displacement_operators(d):
    r"""
    Returns a dictionary associating $(a, b)$ with $\hat{D}_{a,b}$ for $a, b \in [0, d)$.
    """
    return np.array([displace(d, q, p) for q in range(d) for p in range(d)])

def weyl_heisenberg_frame(fiducial):
    r"""
    Applies the $d^2$ displacement operators to a fiducial ket.
    """
    d = fiducial.shape[0]
    D = displacement_operators(d)
    return (D @ fiducial).T.reshape(d, d**2)/np.sqrt(d)

def weyl_heisenberg_povm(fiducial):
    r"""
    Generates a Weyl-Heisenberg POVM by applying the $d^2$ displacement operators to a
    fiducial state and then, if the fiducial state is a ket $\mid \psi \rangle$, forming the projector $\mid \psi \rangle \langle \psi \mid$, and normalizing by $\frac{1}{d}$.

    Note that if the fiducial state is a density matrix, it may be the case that it is invariant under some displacement operators, in which case you'll run into problems!
    """
    if fiducial.shape[1] != 1:
        d = fiducial.shape[0]
        D = displacement_operators(d)
        return np.array([O @ fiducial @ O.conj().T for O in D])/fiducial.shape[0]
    else:
        return frame_povm(weyl_heisenberg_frame(fiducial))



import numpy as np
import scipy as sc
from scipy.special import hyp2f1
import multiprocessing as mp
from itertools import * 

from scipy.special import gamma
from scipy.stats import dirichlet
from scipy.spatial import ConvexHull
from scipy.spatial import HalfspaceIntersection

import matplotlib.pyplot as plt

import jax
import jax.numpy as jp
from jax.config import config
config.update("jax_enable_x64", True)
config.update('jax_platform_name', 'cpu')

from .sics import *
from .weyl_heisenberg import *
from .random import *

####################################################################################################

def flatten(l):
    return [item for sublist in l for item in sublist]

####################################################################################################

def vol_simplex(d):
    return d/gamma(d**2)

def vol_psic(d):
    return np.sqrt((2*np.pi)**(d*(d-1))/(d**(d**2-2)*(d+1)**(d**2-1)))*\
           np.prod([gamma(i) for i in range(1, d+1)])/gamma(d**2)

####################################################################################################

def __make_surface_constraint__(d):
    @jax.jit
    def constraint(V):
        V = jp.abs(V)
        V = V/jp.sum(V)
        return (V@V - 2/(d*(d+1)))**2
    constraint_jac = jax.jit(jax.jacrev(constraint))
    return constraint, constraint_jac

def sample_qplex_surface_point(d, constraint=None, constraint_jac=None):
    if type(constraint) == type(None):
        constraint, constraint_jac = __make_surface_constraint__(d)
    V = np.random.uniform(size=d**2)
    result = sc.optimize.minimize(constraint, V,\
                              jac=constraint_jac,\
                              tol=1e-16,\
                              options={"disp": False,\
                                       "maxiter": 25})
    if not result.success:
        return sample_qplex_surface_point(d, constraint=constraint, constraint_jac=constraint_jac)
    p = np.abs(result.x)
    p = p/np.sum(p)
    return p

def sample_qplex(d, qplex_pts=None, n_qplex_pts=10000, from_surface=True):
    if type(qplex_pts) == type(None):
        qplex_pts = np.array([[1/d if i == j else  1/(d*(d+1))\
                                for j in range(d**2)]\
                                    for i in range(d**2)])
        if n_qplex_pts == d**2:
            return qplex_pts
    if from_surface:
        constraint, constraint_jac = __make_surface_constraint__(d)
    while len(qplex_pts) < n_qplex_pts:
        if from_surface:
            pt = sample_qplex_surface_point(d, constraint=constraint, constraint_jac=constraint_jac)
            inner_products = qplex_pts @ pt
            if np.all(inner_products >= 1/(d*(d+1))):
                qplex_pts = np.vstack([qplex_pts, pt])
        else:
            pt = np.random.dirichlet((1,)*d**2)
            inner_products = qplex_pts @ pt
            if np.all(inner_products >= 1/(d*(d+1)-1e-16)) \
               and np.all(inner_products <= 2/(d*(d+1))+1e-16) \
               and pt @ pt <= 2/(d*(d+1))+1e-16:
                qplex_pts = np.vstack([qplex_pts, pt])
    return qplex_pts

####################################################################################################

def sample_hilbert_qplex(d, n_qplex_pts=10000, from_surface=True):
    qplex_pts = np.array([[1/d if i == j else  1/(d*(d+1)) for j in range(d**2)] for i in range(d**2)])
    if n_qplex_pts == d**2:
        return qplex_pts    
    if from_surface:
        fiducial = sic_fiducial(d)[:,0]
        R = np.array([displace(d, i, j) @ fiducial/np.sqrt(d) for j in range(d) for i in range(d)]).conj()
        return np.vstack([qplex_pts, np.array([abs(R @ rand_ket(d)[:,0])**2 for i in range(n_qplex_pts-d**2)])])
    else:
        R = sic_povm(d)
        return np.vstack([qplex_pts, np.array([R << rand_dm(d) for i in range(n_qplex_pts-d**2)])])

####################################################################################################

def mc_batch(qplex_pts, batch):
    d = int(np.sqrt(qplex_pts.shape[1]))
    hits = 0
    np.random.seed()
    mc_pts = np.random.dirichlet((1,)*d**2, size=batch)
    whittled = mc_pts[np.linalg.norm(mc_pts, axis=1)**2 <= 2/(d*(d+1))]
    if len(whittled) != 0:
        C = np.apply_along_axis(np.all, 1, whittled @ qplex_pts.T >= 1/(d*(d+1)))
        D = dict(zip(*np.unique(C, return_counts=True)))
        hits += D[True] if True in D else 0
    return hits

def cvx_batch(qplex_pts, batch):
    d = int(np.sqrt(qplex_pts.shape[1]))
    hits = 0
    np.random.seed()
    mc_pts = np.random.dirichlet((1,)*d**2, size=batch)
    C = np.array([in_hull(qplex_pts, pt) for pt in mc_pts])
    D = dict(zip(*np.unique(C, return_counts=True)))
    hits += D[True] if True in D else 0
    return hits

def mc_qplex_vol(qplex_pts, n_mc_pts=50000, batch_size=5000, batch_func=None):
    batch_func = mc_batch if batch_func == None else batch_func
    d = int(np.sqrt(qplex_pts.shape[1]))
    n_batches = n_mc_pts // batch_size
    batches = [batch_size]*n_batches
    remaining = n_mc_pts - batch_size*n_batches
    if remaining != 0:
        batches.append(remaining)
    pool = mp.Pool(mp.cpu_count())
    future_res = [pool.apply_async(batch_func, (qplex_pts, batch)) for batch in batches]
    hits = sum([f.get() for f in future_res])
    pool.close()
    return vol_simplex(d)*hits/n_mc_pts

def mc_qplex_vol_error_bars(Q, n_mc_pts=100000, batch_size=5000, N=1000, batch_func=None):
    V = [mc_qplex_vol(Q, n_mc_pts=n_mc_pts, batch_size=batch_size, batch_func=batch_func) for n in range(N)]
    return (np.mean(V), np.std(V))

####################################################################################################

def cvx_qplex_vol(qplex_pts):
    n = qplex_pts.shape[1]
    I = np.eye(n) - np.ones(n)/n
    U, D, V = np.linalg.svd(I@I.T)
    k = np.count_nonzero(np.round(D, decimals=4))
    O = (np.sqrt(np.diag(D[:k])) @ V[:k]).T
    P = sum([np.outer(O[i], I[i]) for i in range(n)])
    return ConvexHull(np.einsum('...ij,...j', P, qplex_pts)).volume

####################################################################################################

def tconstraint(d, t):
    return hyp2f1(1, -t, d, -1)/(d*(d+1))**t

def tcompare(Q, t):
    return (tconstraint(int(np.sqrt(Q.shape[1])), t), np.sum((Q @ Q.T)**t, axis=0)/len(Q))

####################################################################################################

def consistent(Q):
    d = int(np.sqrt(Q.shape[1]))
    L = (1/(d*(d+1)))
    U = (2/(d*(d+1)))
    I = Q @ Q.T
    return np.all(I >= L-1e-16) and np.all(I <= U+1e-16) and \
           np.allclose(np.sum(Q, axis=1), np.ones(Q.shape[0])) and \
           np.all(Q >= 0) and np.all(Q <=1)

####################################################################################################

def qplex_halfspace_intersection(points):
    n = points.shape[0]
    d = int(np.sqrt(points.shape[1]))
    L = (1/(d*(d+1)))
    U = (2/(d*(d+1)))
    halfspaces = np.vstack([np.hstack([-points, L*np.ones(n).reshape(n,1)]),
                            np.hstack([points, -U*np.ones(n).reshape(n,1)])])
    return HalfspaceIntersection(halfspaces, interior_point(halfspaces))

####################################################################################################

def john_plex():
    d = 3
    n = d**2
    L = (1/(d*(d+1)))
    U = (2/(d*(d+1)))
    SIC = (1/d-L)*jp.eye(n) + L*jp.ones((n,n))
    MUBs = np.array(list(set(permutations([0]*3 + [1]*6))))/6
    return np.vstack([SIC, MUBs])

####################################################################################################

from scipy.optimize import linprog

def in_hull(points, x):
    n_points = len(points)
    n_dim = len(x)
    c = np.zeros(n_points)
    A = np.r_[points.T,np.ones((1,n_points))]
    b = np.r_[x, np.ones(1)]
    lp = linprog(c, A_eq=A, b_eq=b)
    return lp.success

####################################################################################################

def add_new_point_on_outsphere(qplex_pts, N=1):
    d = int(np.sqrt(qplex_pts.shape[1]))
    n = d**2
    L = (1/(d*(d+1)))-1e-16
    U = (2/(d*(d+1)))+1e-16

    @jax.jit
    def novelty(V):
        P = V.reshape(N, n)
        P = (P.T/jp.sum(P,axis=1)).T
        return -jp.sum(jp.linalg.norm(qplex_pts - P, axis=1))

    @jax.jit
    def sphere(V):
        P = V.reshape(N, n)
        P = (P.T/jp.sum(P,axis=1)).T
        return (jp.linalg.norm(jp.diag(P @ P.T) - U*jp.ones(N))) #/ 10000000

    @jax.jit
    def consistent_min(V):
        P = V.reshape(N, n)
        P = (P.T/jp.sum(P,axis=1)).T
        Q = jp.vstack([qplex_pts, P])
        return (Q @ Q.T).flatten() - L

    @jax.jit
    def consistent_max(V):
        P = V.reshape(N, n)
        P = (P.T/jp.sum(P,axis=1)).T
        Q = jp.vstack([qplex_pts, P])
        return -(Q @ Q.T).flatten() + U

    result = sc.optimize.minimize(sphere, np.random.randn(N*n),\
                        jac=jax.jit(jax.jacrev(sphere)),\
                        bounds=[(0,1)]*(N*n),
                        tol=1e-16,\
                        constraints=[{"type": "eq",\
                                    "fun": novelty,\
                                    "jac": jax.jit(jax.jacrev(novelty))},
                                    {"type": "ineq",\
                                    "fun": consistent_min,\
                                    "jac": jax.jit(jax.jacrev(consistent_min))},\
                                    {"type": "ineq",\
                                    "fun": consistent_max,\
                                    "jac": jax.jit(jax.jacrev(consistent_max))}],\
                        options={"maxiter": 1000, "disp": True})
    P = jp.abs(result.x.reshape(N, n))
    P = (P.T/np.sum(P,axis=1)).T
    return P

def add_if_consistent(qplex_pts, pts):
    d = int(np.sqrt(qplex_pts.shape[1]))
    L = (1/(d*(d+1)))-1e-16
    U = (2/(d*(d+1)))+1e-16
    J = qplex_pts[:]
    for p in pts:
        I = qplex_pts @ p
        if np.all(I > L) and np.all(I < U):
            J = np.vstack([J, p])
    return J

def permutations_of(v):
    P = np.array(list(set(permutations(flatten([[i]*v[i][1] for i in range(len(v))])))), dtype=np.float128)
    for i, _ in enumerate(v):
        np.place(P, P==i, _[0])
    return P

def add_if_consistent(Q, pts):
    d = int(np.sqrt(Q.shape[1]))
    L = (1/(d*(d+1)))
    U = (2/(d*(d+1)))
    epsilon = 1e-16
    for p in pts:
        I = Q @ p
        if np.all(I >= L-epsilon) and np.all(I <= U+epsilon):
            Q = np.vstack([Q, p])
    return Q

def add_if_consistent_recursive(V):
    def __add_if_helper__(A, B):
        if len(B) == 1:
            return add_if_consistent(A, B[0])
        return __add_if_helper__(__add_if_helper__(A, B[:1]), B[1:])
    return __add_if_helper__(V[0], V[1:])

def non_increasing(a):
    return np.all(np.diff(a) <= 0)


import numpy as np
import scipy as sc
from scipy.stats import unitary_group
from itertools import combinations, permutations

from .utils import *

def random_grassmannian(k, n):
    return unitary_group.rvs(n)[:k]

def standard_grassmannian_form(G):
    return np.linalg.inv(G[:,:2]) @ G

def plucker_coordinate(I, G):
    return np.linalg.det(G[:, I])

def plucker_indices(k, n):
    return list(combinations(list(range(n)), k))

def plucker_coordinates(G):
    return np.array([plucker_coordinate(i, G) for i in plucker_indices(*G.shape)])

###

def __antisymmetrize__(a, b):
    return np.kron(a, b) - np.kron(b, a)

def antisymmetrize(*V):
    return reduce(__antisymmetrize__, V)

###

def plucker_basis(k, n):
    return np.array([antisymmetrize(*[basis(n, i) for i in I]) for I in plucker_indices(k, n)])

def plucker_inner_product(v, w):
    return np.linalg.det(v.conj() @ w.T)

def kplane_inner_product(v, w):
    return abs(plucker_inner_product(v,w))/\
                (np.sqrt(plucker_inner_product(v,v))*\
                 np.sqrt(plucker_inner_product(w,w)))


import numpy as np
import scipy as sc 

import jax
import jax.numpy as jp
from jax.config import config
config.update('jax_platform_name', 'cpu')
config.update("jax_enable_x64", True)

def meff(d, r, max_iter=1000, tol=1e-16,\
                             options={"disp": True,\
                                      "maxiter": 100000},\
                             method="SLSQP"):
    n = d**2
    G_MEFF = np.array([[r if i == j else r*(d*r -1)/(d**2 - 1) for j in range(n)] for i in range(n)])

    @jax.jit
    def objective(V):
        U = (V[:n*d*r] + 1j*V[n*d*r:]).reshape(n, d, r)
        G = jp.einsum("aij, akj, bkl, bil -> ab", U.conj(), U, U.conj(), U).real
        return jp.linalg.norm(G - G_MEFF) + jp.linalg.norm(jp.einsum("...ij,...ij", U.conj(), U) - jp.diag(G))

    for t in range(max_iter):
        result = sc.optimize.minimize(\
                    objective, np.random.randn(2*n*d*r),\
                    jac=jax.jit(jax.jacrev(objective)),\
                    tol=tol,\
                    options=options,\
                    method=method)
        if not np.isclose(result.fun, float("nan"), equal_nan=True):           
            U = (result.x[:n*d*r] + 1j*result.x[n*d*r:]).reshape(n, d, r)
            R = jp.einsum("...ij,...kj", U.conj(), U)
            return R

def check_meff_wh_cov(M, r, max_iter=1000, tol=1e-16, 
                                options={"disp": False,\
                                         "maxiter": 5000},\
                                method="SLSQP",\
                                return_goodies=True):
    n = M.shape[0]
    d = M.shape[1]
    R = M.reshape(n, n)
    D = displacement_operators(d)

    G_MEFF = np.array([[r if i == j else r*(d*r -1)/(d**2 - 1) for j in range(n)] for i in range(n)])
    g_meff = np.sort(G_MEFF.flatten())

    @jax.jit
    def meff_obj(V):
        U = jp.linalg.qr((V[:d**2] + 1j*V[d**2:2*d**2]).reshape(d,d), mode="complete")[0]
        X = jp.linalg.qr((V[2*d**2:2*d**2 + d*r] + 1j*V[2*d**2 + d*r:]).reshape(d, r))[0]
        S = (U @ D @ X @ X.conj().T @ np.transpose(D, (0,2,1)).conj() @ U.conj().T).reshape(n,n)
        return jp.linalg.norm(jp.sort((R @ S.conj().T).flatten()) - g_meff) 
    
    for t in range(max_iter):
        result = sc.optimize.minimize(\
                        meff_obj, np.random.randn(2*d**2 + 2*d*r),\
                        jac=jax.jit(jax.jacrev(meff_obj)),\
                        method="SLSQP")
        if np.isclose(result.fun, 0, atol=1e-3):
            if return_goodies:
                U = jp.linalg.qr((result.x[:d**2] + 1j*result.x[d**2:2*d**2]).reshape(d,d), mode='complete')[0]
                X = jp.linalg.qr((result.x[2*d**2:2*d**2 + d*r] + 1j*result.x[2*d**2 + d*r:]).reshape(d, r))[0]
                return [U, X @ X.conj().T]
            else:
                return True
    return False


import numpy as np
import scipy as sc
from scipy.spatial import ConvexHull
from scipy.spatial import HalfspaceIntersection
from functools import partial

import jax
import jax.numpy as jp
from jax.config import config
config.update('jax_platform_name', 'cpu')
config.update("jax_enable_x64", True)

import cvxpy as cp

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import mpl_toolkits.mplot3d as a3
import matplotlib.colors as colors

from .utils import *
from .random import *

def gpt_full_rank_decomposition(M):
    Q, R = np.linalg.qr(M.T)
    R *= Q[0,0] 
    Q /= Q[0,0]
    B, C = full_rank_decomposition(Q[:, 1:] @ R[1:, ])
    S = np.hstack([Q[:,0:1], B]).T
    E = np.vstack([R[0:1,:], C]).T
    return E, S 

def dual_halfspace_intersection(points):
    points = np.array([pt for pt in points if not np.allclose(pt, np.zeros(points.shape[1]))])
    n = len(points)
    halfspaces = np.vstack([np.hstack([-points, np.zeros(n).reshape(n,1)]),
                            np.hstack([points, -np.ones(n).reshape(n,1)])])
    return HalfspaceIntersection(halfspaces, interior_point(halfspaces))

def plot_convex_hull(hull, points=None, fill=True, ax=None):
    points = hull.points if type(points) == type(None) else points
    d = points.shape[1]

    if type(ax) == type(None):
        fig = plt.figure()
        if d == 3:
            ax = fig.add_subplot(111, projection="3d") 
        else:    
            ax = fig.add_subplot(111)
            ax.set_aspect('equal')

    ax.plot(*[points[:,i] for i in range(d)], 'bo')

    if d == 3:
        if fill:
            c = colors.rgb2hex(sc.rand(3))
            for simplex in hull.simplices:
                f = a3.art3d.Poly3DCollection([[[points[simplex[i], j] for j in range(d)] for i in range(d)]])
                f.set_color(c)
                f.set_edgecolor('k')
                f.set_alpha(0.4)
                ax.add_collection3d(f)
        else:
            for simplex in hull.simplices:
                ax.plot(*[points[simplex, i] for i in range(d)], 'black')
    else:
        for simplex in hull.simplices:
            ax.plot(*[points[simplex, i] for i in range(d)], 'black')
        plt.fill(*[points[hull.vertices,i] for i in range(d)], color=colors.rgb2hex(sc.rand(3)), alpha=0.3)
    return ax

def rank_approximation(P, k, sd=1, tol=1e-6, max_iter=1000, verbose=False):
    m, n = P.shape
    S = np.random.randn(m, k)
    E = np.random.randn(k, n)
    W = np.eye(m*n)/sd**2 
    t = 0; last = None
    while t < max_iter:
        try:
            U = np.kron(E, np.eye(m)) @ W @ np.kron(E.T, np.eye(m))
            V = -2*np.kron(E, np.eye(m)) @ W @ P.T.flatten()
            Y = np.kron(E.T,np.eye(m))

            Svec = cp.Variable(k*m)
            Sprob = cp.Problem(cp.Minimize(cp.quad_form(Svec, U) + V.T @ Svec), [0 <= Y @ Svec, Y @ Svec <= 1])
            Sresult = Sprob.solve()
            S = Svec.value.reshape(k, m).T

            U = np.kron(np.eye(n), S).T @ W @ np.kron(np.eye(n), S)
            V = -2*np.kron(np.eye(n), S).T @ W @ P.T.flatten()
            Y = np.kron(np.eye(n), S)

            Evec = cp.Variable(k*n)
            Eprob = cp.Problem(cp.Minimize(cp.quad_form(Evec, U) + V.T @ Evec), [0 <= Y @ Evec, Y @ Evec <= 1])
            Eresult = Eprob.solve()
            E = Evec.value.reshape(n,k).T
            if verbose and t % 100 == 0:
                print("%d: chi_S = %f | chi_E = %f " % (t, Sresult, Eresult))

            if type(last) != type(None) and abs(Sresult-last[0]) <= tol and abs(Eresult-last[1]) <= tol and not np.isclose(np.sum(S @ E),0):
                break
            last = [Sresult, Eresult]
        except:
            S = np.random.randn(m, k)
            E = np.random.randn(k, n)
            continue
        t += 1
    return S @ E

class GPT:
    @classmethod
    def from_probability_table(cls, P):
        effects, states = gpt_full_rank_decomposition(P)
        return GPT(effects, states)

    def __init__(self, effects, states, unit_effect=None, maximally_mixed_state=None):
        self.unit_effect = unit_effect if type(unit_effect) != type(None) else \
                           np.eye(states.shape[0])[0]
        self.maximally_mixed_state = maximally_mixed_state if type(maximally_mixed_state) != type(None) else \
                                     np.mean(states, axis=1)
        self.effects = np.vstack([effects, self.unit_effect - effects])
        self.states = states

        self.state_space = ConvexHull(self.states.T[:, 1:])
        self.effect_space = ConvexHull(self.effects)

        self.logical_states = dual_halfspace_intersection(self.effects)
        self.logical_effects = dual_halfspace_intersection(self.states.T)

        self.logical_state_space = ConvexHull(self.logical_states.intersections[:, 1:])
        self.logical_effect_space = ConvexHull(self.logical_effects.intersections)
    
        self.d = self.states.shape[0]

    def sample_measurement(self, m, max_iter=100):
        if m == 1:
            return sample_convex_hull(self.effect_space, 1)

        A = self.effect_space.equations[:, :-1]
        b = self.effect_space.equations[:, -1]
        B = np.tile(b, (m,1)).T
        
        @jax.jit
        def unity(V):
            M = V.reshape(m, self.d)
            return jp.linalg.norm(jp.sum(M, axis=0) - self.unit_effect)**2

        @jax.jit
        def consistency(V):
            M = V.reshape(m, self.d)
            return (- A @ M.T - B).flatten()

        i = 0
        while i < max_iter:
            V = np.random.randn(m*self.d)
            result = sc.optimize.minimize(unity, V,\
                                jac=jax.jit(jax.jacrev(unity)),\
                                tol=1e-16,\
                                constraints=[{"type": "ineq",\
                                                "fun": consistency,\
                                                "jac": jax.jit(jax.jacrev(consistency))}],\
                                options={"maxiter": 5000},
                                method="SLSQP")
            M = result.x.reshape(m, self.d)
            if self.valid_measurement(M):
                return M
            i += 1
    
    def valid_measurement(self, M):
        m = M.shape[0]
        A = self.effect_space.equations[:, :-1]
        b = self.effect_space.equations[:, -1]
        B = np.tile(b, (m,1)).T
        return np.all((- A @ M.T - B).flatten() >= 0) and np.allclose(np.sum(M, axis=0), self.unit_effect)

    def sample_logical_measurement(self, m, max_iter=100):
        if m == 1:
            return sample_convex_hull(self.logical_effect_space, 1)

        @jax.jit
        def unity(V):
            M = V.reshape(m, self.d)
            return jp.linalg.norm(jp.sum(M, axis=0) - self.unit_effect)**2

        @jax.jit
        def consistent_min(V):
            M = V.reshape(m, self.d)
            return (M @ self.states).flatten()

        @jax.jit
        def consistent_max(V):
            M = V.reshape(m, self.d)
            return -(M @ self.states).flatten() + 1

        i = 0
        while i < max_iter:
            V = np.random.randn(m*self.d)
            result = sc.optimize.minimize(unity, V,\
                                jac=jax.jit(jax.jacrev(unity)),\
                                tol=1e-16,\
                                constraints=[{"type": "ineq",\
                                             "fun": consistent_min,\
                                             "jac": jax.jit(jax.jacrev(consistent_min))},\
                                            {"type": "ineq",\
                                             "fun": consistent_max,\
                                             "jac": jax.jit(jax.jacrev(consistent_max))}],\
                                options={"maxiter": 5000},
                                method="SLSQP")
            M = result.x.reshape(m, self.d)
            if self.valid_logical_measurement(M):
                return M
            i += 1
    
    def valid_logical_measurement(self, M):
        return np.all(M @ self.states >= 0) and np.all(M @ self.states <= 1) and np.allclose(np.sum(M, axis=0), self.unit_effect)

    def sample_states(self, n):
        return np.vstack([np.ones((1, n)), sample_convex_hull(self.state_space, n).T])

    def sample_logical_states(self, n, max_iter=100):
        @jax.jit
        def info_complete(V):
            S = V.reshape(self.d-1, n)
            S = jp.vstack([jp.ones(n).reshape(1, n), S])
            return (jp.linalg.matrix_rank(S) - self.d).astype(float)**2

        @jax.jit
        def consistent_min(V):
            S = V.reshape(self.d-1, n)
            S = jp.vstack([jp.ones(n).reshape(1, n), S])
            return (self.effects @ S).flatten()

        @jax.jit
        def consistent_max(V):
            S = V.reshape(self.d-1, n)
            S = jp.vstack([jp.ones(n).reshape(1, n), S])
            return -(self.effects @ S).flatten() + 1

        i = 0
        while i < max_iter:
            V = np.random.randn(n*(self.d-1))
            result = sc.optimize.minimize(info_complete, V,\
                                jac=jax.jit(jax.jacrev(info_complete)),\
                                tol=1e-16,\
                                constraints=[{"type": "ineq",\
                                             "fun": consistent_min,\
                                             "jac": jax.jit(jax.jacrev(consistent_min))},\
                                            {"type": "ineq",\
                                             "fun": consistent_max,\
                                             "jac": jax.jit(jax.jacrev(consistent_max))}],\
                                options={"maxiter": 5000},
                                method="SLSQP")
            S = result.x.reshape(self.d-1, n)
            S = np.vstack([np.ones(n).reshape(1,n), S])
            if self.valid_logical_states(S):
                return S
            i += 1

    def valid_states(self, S):
        m = S.shape[1]
        A = self.state_space.equations[:, :-1]
        b = self.state_space.equations[:, -1]
        B = np.tile(b, (m,1)).T
        print(S)
        return np.all((- A @ S[1:, :] - B).flatten() >= 0)
    
    def valid_logical_states(self, S):
        return np.all(self.effects @ S >= 0) and np.all(self.effects @ S <= 1)

class GPTStates:
    def __init__(self, states):
        self.states = states

    def __len__(self):
        return self.states.shape[1]
    
    def dim(self):
        return self.states.shape[0]
    
    def __getitem__(self, key):
        return self.states.T[key]

    def __setitem__(self, key, value):
        self.states.T[key] = value
    
    def __iter__(self):
        return self.states.T.__iter__()

    def __add__(self, other):
        return GPTStates(np.hstack([self.states, other.states]))

    def __mul__(self, other):
        return GPTStates(other*self.states)
    
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        return self.__mul__(1/other)

    def __and__(self, other):
        return GPTStates(kron(self.states, other.states))
    
    def __mod__(self, f):
        return GPTStates(np.array([f(e) for e in self.states.T]).T)

    def __lshift__(self, other):  
        return other @ self.states

    def __rshift__(self, other):
          return other @ np.linalg.pinv(self.states)
    
class GPTEffects:
    def __init__(self, effects):
        self.effects = effects
        self.inverted = False

    def __len__(self):
        return self.effects.shape[0]
    
    def dim(self):
        return self.effects.shape[1]
    
    def __getitem__(self, key):
        return self.effects[key]

    def __setitem__(self, key, value):
        self.effects[key] = value
    
    def __iter__(self):
        return self.effects.__iter__()

    def __add__(self, other):
        return GPTEffects(np.vstack([self.effects, other.effects]))

    def __mul__(self, other):
        return GPTEffects(other*self.effects)
    
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        return self.__mul__(1/other)

    def __and__(self, other):
        return GPTEffects(kron(self.effects, other.effects))
    
    def __mod__(self, f):
        return GPTEffects(np.array([f(e) for e in self.effects]).T)

    def __lshift__(self, other):
        if np.all(other >= 0) and np.all(other <= 1) and np.isclose(np.sum(other),1):
            return np.linalg.pinv(self.effects) @ other
        return self.effects @ other

    def __invert__(self):
        self.inverted = True if not self.inverted else False
        return self

    def __inverted__(self, A):
        A = A if not self.inverted else spectral_inverse(A)
        self.inverted = False if self.inverted else self.inverted
        return A

    def __or__(self, other):
        if type(other) == GPTStates:
            return self.__inverted__(self.effects @ other.states)


import numpy as np
import scipy as sc

from .utils import *

class Operators:
    def __init__(self, E):
        E = E.E if type(E) == Operators else np.array(E)
        if len(E.shape) == 2:
            E = E.reshape(1, *E.shape)
        self.E = E
        self.inverted = False

    @property
    def shape(self):
        return self.E.shape

    def __len__(self):
        r"""
        Number of operators.
        """
        return self.E.shape[0]

    def dim(self):
        r"""
        Dimensionality of the operators.
        """
        return self.E.shape[-1]
        
    def __getitem__(self, key):
        r"""
        Get operator by index.
        """
        return self.E[key]

    def __setitem__(self, key, value):
        r"""
        Set operator by index.
        """
        self.E[key] = value
    
    def __iter__(self):
        r"""
        Iterator over operators.
        """
        return self.E.__iter__()

    def bias(self):
        r"""
        Returns the traces of each operator.
        """
        return np.trace(self.E, axis1=1, axis2=2)
    
    def conj(self):
        r"""
        Conjugate each operator.
        """
        return Operators(self.E.conj())
    
    @property
    def T(self):
        r"""
        Transpose each operator.
        """
        return Operators(np.transpose(self.E, (0,2,1)))

    def __matmul__(self, other):
        r"""
        Pairwise matrix multiplication of two set of of operators.
        """
        return Operators(self.E @ other.E)
    
    def __add__(self, other):
        r"""
        Concatenates to sets of operators.
        """
        return Operators(np.concatenate([self.E, other.E]))

    def __and__(self, other):
        r"""
        Tensor product of elements in self and elements in other.
        """
        return Operators(np.kron(self[:], other[:]))

    def __mul__(self, other):
        r"""
        Scalar multiplication on each operator.
        """
        return Operators(other*self[:])
    
    def __rmul__(self, other):
        r"""
        Scalar multiplication on each operator.
        """
        return self.__mul__(other)
    
    def __truediv__(self, other):
        r"""
        Scalar division on each operator.
        """
        return self.__mul__(1/other)
    
    def upgrade(self, i, dims):
        r"""
        Upgrades operators to act on the i'th place of a tensor product with structure dims.
        """
        return Operators([upgrade(e, i, dims) for e in self[:]])

    def flatten(self):
        r"""
        Returns the matrix whose columns are vectorized operators.
        """
        return self.E.reshape(self.E.shape[0], np.prod(self.E.shape[1:])).T

    def __invert__(self):
        r"""
        Toggles inversion: ~A.
        """
        self.inverted = True if not self.inverted else False
        return self

    def __inverted__(self, A):
        r"""
        Returns A if not inverted else returns the spectral inverse of A, and then uninverts.
        """
        A = A if not self.inverted else spectral_inverse(A)
        self.inverted = False if self.inverted else self.inverted
        return A

    def __xor__(self, other):
        r"""
        A^B gives the (inverted) matrix of traces between elements of self and elements of other.
        """
        return self.__inverted__(self.flatten().conj().T @ other.flatten())
    
    def __or__(self, other):
        r"""
        A|B gives the (inverted) matrix of traces between elements of self and elements of other, the latter normalized to trace 1.
        """
        return self.__inverted__(self.flatten().conj().T @ (other.flatten()/other.bias()))

    def __lshift__(self, other):
        r"""
        A << rho expands rho in terms of A. 
        A << p reconstructs rho from p.
        """
        if len(other.shape) == 1 or other.shape[1] == 1:
            return (self.flatten() @ (~self^self) @ other).reshape(self.E.shape[1:])
        else:
            return (self.flatten().conj().T @ other.flatten()).reshape(len(self), 1)
    
    def superoperator(self):
        r"""
        Returns the superoperator of the elements of self interpreted as a set of Kraus operators.
        """
        return self.__inverted__(sum(kron(e, e.conj()) for e in self[:]))

    def __lt__(self, other):
        r"""
        A < rho applies A to rho, where A is interpreted as a set of Kraus operators forming a channel.
        Similarly A < B applies A to each element of B.
        """
        M = self.superoperator() @ other.flatten()
        d = int(np.sqrt(len(M)))
        return M.reshape(d,d) if type(other) == np.ndarray else \
                 Operators(M.T.reshape(len(other),d,d))
                 
    def __floordiv__(self, other):
        r"""
        Returns the (inverse) frame superoperator where other is the post measurement states.
        """
        return self.__inverted__(self.flatten() @ other.flatten().conj().T)

    def __pow__(self, other):
        r"""
        Returns dual elements with respect to other.
        E.g. if self is POVM and other is post-measurement states, this gives dual POVM elements.
        If self is post-measurement states and other is POVM, this gives dual post measurement elements.
        Together they give the dual reference measurement.
        """
        return Operators(((~self // other) @ self.flatten()).T.reshape(self.E.shape))

    def __mod__(self, f):
        r"""
        Applies function f to each operator.
        """
        return Operators(np.array([f(e) for e in self[:]]))


import numpy as np
from .utils import *
import sympy
import sympy.liealgebras.type_a

def mercedes_benz_frame():
    return np.array([[r.real, r.imag] for r in [np.exp(2j*np.pi*i/3) for i in range(3)]]).T*np.sqrt(2/3)

def circular_shifts(v):
    shifts = [v]
    for i in range(len(v)-1):
        u = shifts[-1][:]
        u.insert(0, u.pop()) 
        shifts.append(u)
    return shifts

def icosahedron_vertices():
    phi = (1+np.sqrt(5))/2
    return [np.array(v) for v in 
               circular_shifts([0, 1, phi]) + \
               circular_shifts([0, -1, -phi]) + \
               circular_shifts([0, 1, -phi]) + \
               circular_shifts([0, -1, phi])]

def icosahedron_frame():
    vertices = icosahedron_vertices()
    keep = []
    for i, a in enumerate(vertices):
        for j, b in enumerate(vertices):
            if i != j and np.allclose(a, -b) and j not in keep:
                keep.append(i)
    vertices = [normalize(e) for i, e in enumerate(vertices) if i in keep]
    return np.array(vertices).T*np.sqrt(3/len(vertices))

def petersen_frame():
    petersen_vertices = ["u1", "u2", "u3", "u4", "u5", "v1", "v2", "v3", "v4", "v5"]
    petersen_graph = \
        {"u1": ["v1", "u2", "u5"],
        "u2": ["u1", "v2", "u3"],
        "u3": ["u2", "v3", "u4"],
        "u4": ["u3", "v4", "u5"],
        "u5": ["u4", "v5", "u1"],
        "v1": ["u1", "v4", "v3"],
        "v2": ["u2", "v4", "v5"],
        "v3": ["v5", "v1", "u3"],
        "v4": ["u4", "v1", "v2"],
        "v5": ["u5", "v3", "v2"]}
    petersen_gram = np.array([[1 if a == b else (\
                               -2/3 if b in petersen_graph[a] else \
                               1/6) for b in petersen_vertices]\
                                        for a in petersen_vertices]) 
    U, D, V = np.linalg.svd(petersen_gram)
    return (V[:4].T @ np.sqrt(np.diag(D[:4]))*np.sqrt(4/10)).T

def honey_frame(d, full=False):
    n = d*(d+1)/2
    A = sympy.liealgebras.type_a.TypeA(d)
    roots = [np.array(root) for root in A.positive_roots().values()]
    if full:
        roots.extend([-r for r in roots])
    U, D, V = np.linalg.svd(np.array(roots).T)
    return np.array([state/np.linalg.norm(state) for state in V[:d].T]).T*np.sqrt(d/n)



import numpy as np
import pkg_resources
from itertools import product

from .weyl_heisenberg import *
from .povm_utils import *

def sic_fiducial(d):
    r"""
    Loads a Weyl-Heisenberg covariant SIC-POVM fiducial state of dimension $d$ from the repository provided here: http://www.physics.umb.edu/Research/QBism/solutions.html.
    """
    f = pkg_resources.resource_stream(__name__, "sic_povms/d%d.txt" % d)
    fiducial = []
    for line in f:
        if line.strip() != "":
            re, im = [float(v) for v in line.split()]
            fiducial.append(re + 1j*im)
    return normalize(np.array(fiducial).reshape(d, 1))

def sic_frame(d):
    r"""
    Returns the $d^2$ states constructed by applying the Weyl-Heisenberg displacement operators to the SIC-POVM fiducial state of dimension $d$.
    """
    return weyl_heisenberg_frame(sic_fiducial(d))

def sic_povm(d):
    r"""
    Returns a SIC-POVM of dimension $d$.
    """
    return frame_povm(weyl_heisenberg_frame(sic_fiducial(d)))

def hoggar_sic_fiducial():
    r"""
    Returns a fiducial state for the exceptional SIC in dimension $8$, the Hoggar SIC.

    Unnormalized: $\begin{pmatrix} -1 + 2i \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \end{pmatrix}$.
    """
    return normalize(np.array([-1 + 2j, 1, 1, 1, 1, 1, 1, 1])).reshape(8, 1)

def hoggar_sic_povm():
    r"""
    Constructs the Hoggar POVM, which is covariant under the tensor product of three copies of the $d=2$ Weyl-Heisenberg group. In other words, we apply the 64 displacement operators:

    $$ \hat{D}_{a, b, c, d, e, f} = X^{a}Z^{b} \otimes X^{c}Z^{d} \otimes X^{e}Z^{f} $$

    to the Hoggar fiducial state, form the corresponding projectors, and rescale by $\frac{1}{8}$.
    """
    d = 8
    Z, X = clock(2), shift(2)
    indices = list(product([0,1], repeat=6))
    D = np.array([kron(np.linalg.matrix_power(X, I[0]) @ np.linalg.matrix_power(Z, I[1]),\
                       np.linalg.matrix_power(X, I[2]) @ np.linalg.matrix_power(Z, I[3]),\
                       np.linalg.matrix_power(X, I[4]) @ np.linalg.matrix_power(Z, I[5])) for I in indices])
    return frame_povm((D @ hoggar_sic_fiducial()).T.reshape(d, d**2)/np.sqrt(d))


import setuptools

setuptools.setup(
    name = "qbuki",
    version = "0.1.7",
    author = "Matthew Weiss",
    author_email = "heyredhat@gmail.com",
    description = "QBism for Python",
    url = "https://github.com/heyredhat/qbuki",
    packages = setuptools.find_packages(),
    include_package_data = True)


import numpy as np
import scipy as sc
from functools import reduce

def normalize(v):
    r"""
    Normalizes vector.
    """
    return v/np.linalg.norm(v)

def basis(d, i):
    r"""
    Returns basis vectors.
    """
    return np.eye(d)[i]

def kron(*args):
    r"""
    Numpy's kronecker product with a variable number of arguments.
    """
    return reduce(lambda x, y: np.kron(x, y), args)

def upgrade(O, i, dims):
    return kron(*[O if i == j else np.eye(d) for j, d in enumerate(dims)])

def pnorm(A, p=2):
    S = np.linalg.svd(A, compute_uv=False)
    return np.sum(S**p)**(1/p) if p != np.inf else np.max(S)

def paulis():
    r"""
    Returns Pauli matrices.
    """
    X = np.array([[0, 1], [1, 0]])
    Y = np.array([[0, -1j], [1j, 0]])
    Z = np.array([[1, 0], [0, -1]])
    return np.array([np.eye(2), X, Y, Z])

def gelmann_basis(d):
    r"""
    Hermitian operator basis.
    """
    def h_helper(d,k):
        if k == 1:
            return np.eye(d)
        if k > 1 and k < d:
            return sc.linalg.block_diag(h_helper(d-1, k), [0])
        if k == d:
            return np.sqrt(2/(d*(d+1)))*sc.linalg.block_diag(np.eye(d-1), [1-d])

    E = [[np.zeros((d,d)) for k in range(d)] for j in range(d)]
    for j in range(d):
        for k in range(d):
            E[j][k][j,k] = 1
    F = []
    for j in range(d):
        for k in range(d):
            if k < j:
                F.append(E[k][j] + E[j][k])
            elif k > j:
                F.append(-1j*(E[j][k] - E[k][j]))
    F.extend([h_helper(d, k) for k in range(1,d+1)])
    return np.array([f/np.sqrt((f@f).trace()) for f in F])

def fft_matrix(d):
    r"""
    Finite fourier transform matrix.
    """
    w = np.exp(2*np.pi*1j/d)
    return np.array([[w**(i*j) for j in range(d)] for i in range(d)])/np.sqrt(d)

def full_rank_decomposition(M):
    U, D, V = np.linalg.svd(M)
    m = (np.isclose(D, 0)).argmax()
    m = m if m != 0 else len(D)
    sqrtD = np.sqrt(np.diag(D[:m]))
    return U[:,:m] @ sqrtD, sqrtD @ V[:m,:]

def spectral_inverse(M):
    r"""
    Spectral/group inverse.
    """
    B, C = full_rank_decomposition(M)
    return (B @ np.linalg.matrix_power(C @ B, -2) @ C)

def partial_trace_kraus(keep, dims):
    r"""
    Constructs the Kraus map corresponding to the partial trace. Takes `keep` which is a single index or list of indices denoting
    subsystems to keep, and a list `dims` of dimensions of the overall tensor product Hilbert space. 

    For illustration, to trace over the $i^{th}$ subsystem of $n$, one would construct Kraus operators:

    $$ \hat{K}_{i} = I^{\otimes i - 1} \otimes \langle i \mid \otimes I^{\otimes n - i}$$.
    """
    if type(keep) == int:
        keep = [keep]
    trace_over = [i for i in range(len(dims)) if i not in keep]
    indices = [{trace_over[0]:t} for t in range(dims[trace_over[0]])]
    for i in trace_over[1:]:
        new_indices = []
        for t in range(dims[i]):
            new_indices.extend([{**j, **{i: t}} for j in indices])
        indices = new_indices
    return np.array([kron(*[np.eye(d) if i in keep else basis(d, index[i]) for i, d in enumerate(dims)]) for index in indices])

def state_space_dimension(d, field):
    r"""
    Returns the dimension of the state space for a give pure state space dimension and number field.
    """
    if field == "complex":
        n = int(d**2)
    elif field == "real":
        n = int(d*(d+1)/2)
    return n

def stereographic_projection(X, pole=None):
    if type(pole) == type(None):
        pole = np.eye(len(X))[-1]
    if np.isclose(X@pole, 1):
        return np.inf
    return (pole + (X - pole)/(1-X@pole))[:-1]

def reverse_stereographic_projection(X, pole=None):
    if type(X) != np.ndarray and X == np.inf:
        return pole
    if type(pole) == type(None):
        pole = np.eye(len(X)+1)[0]
    X = np.append(X, 0)
    return ((np.linalg.norm(X)**2 - 1)/(np.linalg.norm(X)**2 +1))*pole + (2/(np.linalg.norm(X)**2 +1))*X

def interior_point(halfspaces):
    norm_vector = np.reshape(np.linalg.norm(halfspaces[:, :-1], axis=1), (halfspaces.shape[0], 1))
    c = np.zeros((halfspaces.shape[1],)); c[-1] = -1
    A = np.hstack((halfspaces[:, :-1], norm_vector))
    b = -halfspaces[:, -1:]
    res = sc.optimize.linprog(c, A_ub=A, b_ub=b, bounds=(None, None))
    return res.x[:-1]

def gram_to_vecs(G):
    """
    Reconstructs vectors from the little Gram matrix via SVD.
    """
    n = G.shape[0]
    U, D, V = np.linalg.svd(G)
    d = np.count_nonzero(np.round(D, decimals=4))
    return (np.sqrt(np.diag(D[:d])) @ V[:d])

def permuter(perm, d):
    return sum([np.outer(kron(*[basis(d, j) for j in np.array(prod)[tuple(perm),]]),\
                         kron(*[basis(d, i) for i in prod])) for prod in product(list(range(d)), repeat=len(perm))])
        
def cntrl_permuter(perm, d):
    return sc.linalg.block_diag(np.eye(d**len(perm)), permuter(perm, d))